#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""`SPXtlm`, class to read/access PACE/SPEXone telemetry data."""

from __future__ import annotations

__all__ = ["SPXtlm"]

import datetime as dt
import logging
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np

from .hkt_io import HKTio
from .lib import pyspex_version
from .lib.hk_tlm import HKtlm
from .lib.leap_sec import get_leap_seconds
from .lib.sci_tlm import DET_CONSTS, SCItlm
from .lv0_lib import dump_hkt, dump_science, read_lv0_data

if TYPE_CHECKING:
    from dataclasses import dataclass

    import numpy.typing as npt


# - global parameters -----------------------
module_logger = logging.getLogger("pyspex.tlm")

FULLFRAME_BYTES = 2 * DET_CONSTS["dimFullFrame"]
DATE_MIN = dt.datetime(2020, 1, 1, 1, tzinfo=dt.UTC)
TSTAMP_MIN = int(DATE_MIN.timestamp())


# - helper functions ------------------------
def get_l1a_filename(
    config: dataclass,
    coverage: tuple[dt.datetime, dt.datetime],
    mode: str | None = None,
) -> Path:
    """Return filename of level-1A product.

    Parameters
    ----------
    config :  dataclass
        Settings for the L0->l1A processing
    coverage :  tuple[dt.datetime, dt.datetime]
        time_coverage_start and time_coverage_end of the data
    mode :  {'binned', 'full'} | None, default=None
        Select Science packages with full-frame images or binned images

    Returns
    -------
    Path
        Filename of level-1A product.

    Notes
    -----
    === Inflight ===
    L1A file name format, following the NASA ... naming convention:
        PACE_SPEXONE[_TTT].YYYYMMDDTHHMMSS.L1A[.Vnn].nc
    where
        TTT is an optional data type (e.g., for the calibration data files)
        YYYYMMDDTHHMMSS is time stamp of the first image in the file
        Vnn file-version number (omitted when nn=1)

    for example (file-version=1):
        [Science Product] PACE_SPEXONE.20230115T123456.L1A.nc
        [Calibration Product] PACE_SPEXONE_CAL.20230115T123456.L1A.nc
        [Dark science Product] PACE_SPEXONE_DARK.20230115T123456.L1A.nc

    === OCAL ===
    L1A file name format:
        SPX1_OCAL_<msm_id>[_YYYYMMDDTHHMMSS]_L1A_vvvvvvv.nc
    where
        msm_id is the measurement identifier
        YYYYMMDDTHHMMSS is time stamp of the first image in the file
        vvvvvvv is the git-hash string of the pyspex repository

    """
    if config.outfile:
        return config.outdir / config.outfile

    # +++++ in-flight product-name convention +++++
    if config.l0_format != "raw":
        subtype = ""
        if config.eclipse is None:
            subtype = "_OCAL"
        elif config.eclipse:
            subtype = "_CAL" if mode == "full" else "_DARK"

        prod_ver = (
            "" if config.processing_version == 1 else f".V{config.processing_version:d}"
        )
        return config.outdir / (
            f"PACE_SPEXONE{subtype}"
            f".{coverage[0].strftime('%Y%m%dT%H%M%S'):15s}"
            f".L1A{prod_ver}.nc"
        )

    # +++++ OCAL product-name convention +++++
    # determine measurement identifier
    msm_id = config.l0_list[0].stem
    try:
        new_date = dt.datetime.strptime(msm_id[-22:], "%y-%j-%H:%M:%S.%f").strftime(
            "%Y%m%dT%H%M%S.%f"
        )
    except ValueError:
        pass
    else:
        msm_id = msm_id[:-22] + new_date

    return config.outdir / f"SPX1_OCAL_{msm_id}_L1A_{pyspex_version(githash=True)}.nc"


# - class SPXtlm ----------------------------
class SPXtlm:
    """Access/convert parameters of SPEXone Science telemetry data.

    Notes
    -----
    This class has the following methods::

     - set_coverage(coverage: tuple[datetime, datetime] | None,
                    extent: bool = False) -> None
     - reference_date() -> datetime
     - time_coverage_start() -> datetime
     - time_coverage_end() -> datetime
     - from_hkt(flnames: str | Path | list[Path], *,
                instrument: str | None = None, dump: bool = False) -> None
     - from_lv0(flnames: str | Path | list[Path], file_format: str = 'dsb',
                *, tlm_type: str | None = None,
                debug: bool = False, dump: bool = False) -> None
     - from_l1a(flname: str | Path,
                *, tlm_type: str | None = None,
                mps_id: int | None = None) -> None


    Examples
    --------
    Read data from SPEXone level-0 products

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_lv0(list_l0_products, tlm_type='hk')
    # returns list of TcAccept, TcReject, TcExecute, TcFail and EventRp
    >>> spx.nomhk.events
    # return detector images
    >>> spx.from_lv0(list_l0_products, tlm_type='sci')
    >>> spx.science.images

    Read data from a SPEXone level-1a product

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_l1a(l1a_product, tlm_type='sci', mps_id=47)
    >>> spx.science.images

    Read data from a PACE-HKT product

    >>> from pyspex.tlm import SPXtlm
    >>> spx = SPXtlm()
    >>> spx.from_hkt(hkt_product)
    # returns list of TcAccept, TcReject, TcExecute, TcFail and EventRp
    >>> spx.nomhk.events

    """

    def __init__(self: SPXtlm) -> None:
        """Initialize SPXtlm object."""
        self.mode = "all"
        self.logger = logging.getLogger(__name__)
        self.file_list: list[Path] | None = None
        self._coverage: list[dt.datetime, dt.datetime] | None = None
        self.nomhk: HKtlm = HKtlm()
        self.science: SCItlm = SCItlm()

    @property
    def coverage(self: SPXtlm) -> list[dt.datetime, dt.datetime] | None:
        """Return data time_coverage."""
        return self._coverage

    def set_coverage(
        self: SPXtlm,
        coverage_new: tuple[dt.datetime, dt.datetime] | None,
        extent: bool = False,
    ) -> None:
        """Set, reset or update the class attribute `coverage`.

        Parameters
        ----------
        coverage_new :  list[dt.datetime, dt.datetime] | None,
           Provide new time_coverage range, or reset when coverage_new is None
        extent :  bool, default=False
           Extent existing time_coverage range with coverage_new

        Examples
        --------
        Reset self.coverage to a new time_coverage:
        > self.set_coverage(None)
        > self.set_coverage(coverage_new)

        Extent self.coverage with a new time_coverage
        if self.coverage is not None else self.coverage will be coverage_new:
        > self.set_coverage(coverage_new, extent=True)

        This will not change self.coverage to coverage_new
        when self.coverage is not None:
        > self.set_coverage(coverage_new)

        """
        if self._coverage is None or coverage_new is None:
            self._coverage = coverage_new
            return

        if not extent:
            return

        one_hour = dt.timedelta(hours=1)
        if self._coverage[0] - one_hour < coverage_new[0] < self._coverage[0]:
            self._coverage[0] = coverage_new[0]
        if self._coverage[1] < coverage_new[1] < self._coverage[1] + one_hour:
            self._coverage[1] = coverage_new[1]

    @property
    def reference_date(self: SPXtlm) -> dt.datetime:
        """Return date of reference day (tzone aware)."""
        if self._coverage is None:
            raise ValueError("no valid timestamps found")

        return dt.datetime.combine(self._coverage[0].date(), dt.time(0), dt.UTC)

    @property
    def time_coverage_start(self: SPXtlm) -> dt.datetime | None:
        """Return time_coverage_start."""
        return None if self._coverage is None else self._coverage[0]

    @property
    def time_coverage_end(self: SPXtlm) -> dt.datetime | None:
        """Return time_coverage_end."""
        return None if self._coverage is None else self._coverage[1]

    def sel(self: SPXtlm, mask: npt.NDArray[bool]) -> SPXtlm:
        """Return subset of SPXtlm object using a mask array."""
        spx = copy(self)
        spx.set_coverage(None)
        spx.science = self.science.sel(mask)

        # set Science time_coverage_range
        indices = mask.nonzero()[0]
        master_cycle = dt.timedelta(milliseconds=self.science.master_cycle(-1))
        spx.set_coverage(None)
        if len(indices) == 1:
            spx.set_coverage(
                (
                    spx.science.tstamp[0]["dt"],
                    spx.science.tstamp[0]["dt"] + master_cycle,
                )
            )
        else:
            spx.set_coverage(
                (
                    spx.science.tstamp[0]["dt"],
                    spx.science.tstamp[-1]["dt"] + master_cycle,
                )
            )
        # select nomhk data within Science time_coverage_range
        hk_tstamps = np.array(
            [x.replace(tzinfo=None) for x in self.nomhk.tstamp], dtype="datetime64"
        )
        dt_min = np.datetime64(spx.coverage[0].replace(tzinfo=None)) - np.timedelta64(
            1, "s"
        )
        dt_max = np.datetime64(spx.coverage[1].replace(tzinfo=None)) + np.timedelta64(
            1, "s"
        )
        spx.nomhk = self.nomhk.sel((hk_tstamps >= dt_min) & (hk_tstamps <= dt_max))
        return spx

    def from_hkt(
        self: SPXtlm,
        flnames: str | Path | list[Path],
        *,
        instrument: str | None = None,
        dump: bool = False,
    ) -> None:
        """Read telemetry data from PACE HKT product(s).

        Parameters
        ----------
        flnames :  str | Path | list[Path]
           sorted list of PACE_HKT filenames (netCDF4 format)
        instrument :  {'spx', 'sc', 'oci', 'harp'}, default='spx'
           abbreviations for the PACE instruments
        dump :  bool, default=False
           dump header information of the telemetry packages @1Hz for
           debugging purposes

        """
        self.set_coverage(None)
        if isinstance(flnames, str | Path):
            flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        if instrument is None:
            instrument = "spx"
        elif instrument not in ["spx", "sc", "oci", "harp"]:
            raise KeyError("instrument not in ['spx', 'sc', 'oci', 'harp']")

        self.file_list = flnames

        # check number of telemetry data-packages
        ccsds_hk = HKTio(flnames).housekeeping(instrument)
        if not ccsds_hk:
            return

        # perform dump of telemetry data-packages
        if dump:
            dump_hkt(flnames[0].stem + "_hkt.dump", ccsds_hk)
            return

        # check if TAI timestamp is valid
        ii = len(ccsds_hk) // 2
        if (tai_sec := ccsds_hk[ii]["hdr"]["tai_sec"][0]) == 0:
            return

        # set epoch
        epoch = dt.datetime(1958, 1, 1, tzinfo=dt.UTC)
        epoch -= dt.timedelta(seconds=get_leap_seconds(float(tai_sec)))
        self.nomhk.extract_l0_hk(ccsds_hk, epoch)

        # reject nomHK records before or after a big time-jump
        _mm = np.diff(self.nomhk.tstamp) > dt.timedelta(days=1)
        if np.any(_mm):
            indx = _mm.nonzero()[0]
            _mm = np.full(self.nomhk.size, True, dtype=bool)
            _mm[indx[0] + 1 :] = False
            if np.sum(_mm) < self.nomhk.size // 2:
                _mm = ~_mm
            self.logger.warning(
                "rejected nomHK: %d -> %d", self.nomhk.size, np.sum(_mm)
            )
            self.nomhk.sel(_mm)

        # set time-coverage
        self.set_coverage(
            (self.nomhk.tstamp[0], self.nomhk.tstamp[-1] + dt.timedelta(seconds=1))
        )

    def from_lv0(
        self: SPXtlm,
        flnames: str | Path | list[Path],
        file_format: str = "dsb",
        *,
        tlm_type: str | None = None,
        debug: bool = False,
        dump: bool = False,
    ) -> None:
        """Read telemetry data from SPEXone level-0 product(s).

        Parameters
        ----------
        flnames :  str | Path | list[Path]
           sorted list of CCSDS filenames
        file_format : {'raw', 'st3', 'dsb'}, default='dsb'
           type of CCSDS data
        tlm_type :  {'hk', 'sci', 'all'}, default='all'
           select type of telemetry packages
           Note that we always read the complete level-0 products.
        debug : bool, default=False
           run in debug mode, read only packages heades
        dump :  bool, default=False
           dump header information of the telemetry packages @1Hz for
           debugging purposes

        """
        self.set_coverage(None)
        if isinstance(flnames, str | Path):
            flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        if file_format not in ["raw", "st3", "dsb"]:
            raise KeyError("file_format not in ['raw', 'st3', 'dsb']")
        if tlm_type is None:
            tlm_type = "all"
        elif tlm_type not in ["hk", "sci", "all"]:
            raise KeyError("tlm_type not in ['hk', 'sci', 'all']")
        self.file_list = flnames

        # read telemetry data
        ccsds_sci, ccsds_hk = read_lv0_data(flnames, file_format, debug=debug)

        # perform ASCII dump
        if dump and ccsds_hk:
            dump_hkt(flnames[0].stem + "_hkt.dump", ccsds_hk)
        if dump and ccsds_sci:
            dump_science(flnames[0].stem + "_sci.dump", ccsds_sci)

        # exit when debugging or only an ASCII data-dump is requested
        if debug or dump:
            return

        # exit when Science data is requested and no Science data is available
        #   or when housekeeping is requested and no housekeeping data is available
        if (not ccsds_sci and tlm_type == "sci") or (not ccsds_hk and tlm_type == "hk"):
            self.logger.info("Asked for tlm_type=%s, but none found", tlm_type)
            return

        # set epoch
        if file_format == "dsb":
            # check if TAI timestamp is valid
            ii = len(ccsds_hk) // 2
            if (tai_sec := ccsds_hk[ii]["hdr"]["tai_sec"][0]) == 0:
                return

            epoch = dt.datetime(1958, 1, 1, tzinfo=dt.UTC)
            epoch -= dt.timedelta(seconds=get_leap_seconds(float(tai_sec)))
        else:
            epoch = dt.datetime(1970, 1, 1, tzinfo=dt.UTC)

        if tlm_type != "hk":
            # collect Science telemetry data
            if self.science.extract_l0_sci(ccsds_sci, epoch) == 0:
                self.logger.info("no valid Science package found")
            else:
                # reject Science records before or after a big time-jump
                _mm = np.diff(self.science.tstamp["dt"]) > dt.timedelta(days=1)
                if np.any(_mm):
                    indx = _mm.nonzero()[0]
                    _mm = np.full(self.science.size, True, dtype=bool)
                    _mm[indx[0] + 1 :] = False
                    if np.sum(_mm) < self.science.size // 2:
                        _mm = ~_mm
                    self.logger.warning(
                        "rejected science: %d -> %d", self.science.size, np.sum(_mm)
                    )
                    self.science.sel(_mm)

                # set time-coverage
                master_cycle = dt.timedelta(milliseconds=self.science.master_cycle(-1))
                self.set_coverage(
                    (
                        self.science.tstamp["dt"][0],
                        self.science.tstamp["dt"][-1] + master_cycle,
                    )
                )

        # collected NomHK telemetry data
        if tlm_type != "sci" and ccsds_hk:
            self.nomhk.extract_l0_hk(ccsds_hk, epoch)
            if self.nomhk.size == 0:
                self.logger.info("no valid housekeeping package found")
                return

            # reject nomHK records before or after a big time-jump
            _mm = np.diff(self.nomhk.tstamp) > dt.timedelta(days=1)
            if np.any(_mm):
                indx = _mm.nonzero()[0]
                _mm = np.full(self.nomhk.size, True, dtype=bool)
                _mm[indx[0] + 1 :] = False
                if self.coverage is None:
                    if np.sum(_mm) < self.nomhk.size // 2:
                        _mm = ~_mm
                else:
                    dt_start = abs(self.coverage[0] - self.nomhk.tstamp[0])
                    dt_end = abs(self.coverage[-1] - self.nomhk.tstamp[-1])
                    if dt_start > dt_end:
                        _mm = ~_mm
                self.logger.warning(
                    "rejected nomHK: %d -> %d", self.nomhk.size, np.sum(_mm)
                )
                self.nomhk.sel(_mm)

            # set time-coverage (only, when self._coverage is None)
            self.set_coverage(
                (self.nomhk.tstamp[0], self.nomhk.tstamp[-1] + dt.timedelta(seconds=1))
            )

    def from_l1a(
        self: SPXtlm,
        flname: str | Path,
        *,
        tlm_type: str | None = None,
        mps_id: int | None = None,
    ) -> None:
        """Read telemetry data from SPEXone level-1A product.

        Parameters
        ----------
        flname :  Path
           name of one SPEXone level-1A product
        tlm_type :  {'hk', 'sci', 'all'}, default='all'
           select type of telemetry packages
        mps_id :  int, optional
           select on MPS ID

        """
        self.set_coverage(None)
        if tlm_type is None:
            tlm_type = "all"
        elif tlm_type not in ["hk", "sci", "all"]:
            raise KeyError("tlm_type not in ['hk', 'sci', 'all']")

        self.file_list = [flname]
        with h5py.File(flname) as fid:
            # collect Science telemetry data
            if tlm_type != "hk":
                self.science.extract_l1a_sci(fid, mps_id)
                if self.science.size == 0:
                    return

                _mm = self.science.tstamp["tai_sec"] > TSTAMP_MIN
                if np.any(_mm):
                    tstamp = self.science.tstamp["dt"][_mm]
                    ii = int(np.nonzero(_mm)[0][-1])
                    master_cycle = dt.timedelta(
                        milliseconds=self.science.master_cycle(ii)
                    )
                    self.set_coverage((tstamp[0], tstamp[-1] + master_cycle))

            # collected NomHk telemetry data
            if tlm_type != "sci":
                self.nomhk.extract_l1a_hk(fid, mps_id)
                if self.nomhk.size == 0:
                    return
                self.set_coverage(
                    (
                        self.nomhk.tstamp[0],
                        self.nomhk.tstamp[-1] + dt.timedelta(seconds=1),
                    )
                )

    def full(self: SPXtlm) -> SPXtlm:
        """Select full-frame measurements."""
        self.mode = "full"
        if self.science.size == 0:
            return self

        sci_mask = (self.science.tstamp["tai_sec"] > TSTAMP_MIN) & (
            self.science.tlm["IMRLEN"] == FULLFRAME_BYTES
        )
        if np.all(~sci_mask):
            return SPXtlm()  # return empy object
        if np.all(sci_mask):
            return self  # return original object
        self.logger.debug("Rejected %d binned Science images", np.sum(~sci_mask))

        return copy(self).sel(sci_mask)

    def binned(self: SPXtlm) -> SPXtlm:
        """Select binned images from data."""
        self.mode = "binned"
        if self.science.size == 0:
            return self

        sci_mask = (self.science.tstamp["tai_sec"] > TSTAMP_MIN) & (
            self.science.tlm["IMRLEN"] < FULLFRAME_BYTES
        )
        if np.all(~sci_mask):
            return SPXtlm()  # return empy object
        if np.all(sci_mask):
            return self  # return original object
        self.logger.debug("Rejected %d full-frame Science images", np.sum(~sci_mask))

        return copy(self).sel(sci_mask)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2023-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""`SCItlm`, class to read/access PACE/SPEXone telemetry data."""

from __future__ import annotations

__all__ = ["DET_CONSTS", "SCItlm"]

import datetime as dt
import logging
from copy import copy
from typing import TYPE_CHECKING

import numpy as np

from .ccsds_hdr import CCSDShdr
from .tlm_utils import convert_hk

if TYPE_CHECKING:
    import h5py
    from numpy.typing import ArrayLike, NDArray


# - global parameters -----------------------
DET_CONSTS = {
    "dimRow": 2048,
    "dimColumn": 2048,
    "dimFullFrame": 2048 * 2048,
    "DEM_frequency": 10,  # [MHz]
    "FOT_diagnostic": 0.4644,  # [ms]
    "FTI_diagnostic": 240.0,  # [ms]
    "FTI_margin": 212.4,  # [ms]
    "FOT_science": 0.3096,  # [ms]
    "FTI_science": 1000 / 15,  # [ms]
    "FOT_length": 20,
}
FULLFRAME_BYTES = 2 * DET_CONSTS["dimFullFrame"]

TSTAMP_MIN = 1561939200  # 2019-07-01T00:00:00+00:00
TSTAMP_TYPE = np.dtype([("tai_sec", int), ("sub_sec", int), ("dt", "O")])

module_logger = logging.getLogger("pyspex.lib.sci_tlm")


# - helper functions ------------------------
def subsec2musec(sub_sec: int) -> int:
    """Return subsec as microseconds."""
    return int(1e6 * sub_sec / 65536)


def mask2slice(mask: NDArray[bool]) -> None | slice | tuple | NDArray[bool]:
    """Try to slice (faster), instead of boolean indexing (slow)."""
    if np.all(~mask):
        return None
    if np.all(mask):
        return np.s_[:]  # read everything

    indx = np.nonzero(mask)[0]
    if np.all(np.diff(indx) == 1):
        # perform start-stop indexing
        return np.s_[indx[0] : indx[-1] + 1]

    # perform boolean indexing
    return mask


# - class SCItlm ----------------------------
class SCItlm:
    """Class to handle SPEXone Science-telemetry packets."""

    def __init__(self: SCItlm) -> None:
        """Initialize SCItlm object."""
        self.hdr: NDArray | None = None
        self.tlm: NDArray | None = None
        self.tstamp: NDArray | None = None
        self.images: tuple[NDArray, ...] | tuple[()] = ()

    def init_attrs(self: SCItlm) -> None:
        """Initialize class attributes."""
        self.hdr = None
        self.tlm = None
        self.tstamp = None
        self.images = ()

    @property
    def size(self: SCItlm) -> int:
        """Return number of elements."""
        return 0 if self.tlm is None else len(self.tlm)

    def copy(self: SCItlm) -> SCItlm:
        """Return deep-copy of SCItlm object."""
        sci = SCItlm()
        sci.hdr = self.hdr.copy()
        sci.tlm = self.tlm.copy()
        sci.tstamp = self.tstamp.copy()
        sci.images = copy(self.images)
        return sci

    def sel(self: SCItlm, mask: ArrayLike[bool]) -> SCItlm:
        """Return subset of SCItlm object using a mask array."""
        sci = SCItlm()
        if self.hdr is not None:
            sci.hdr = self.hdr[mask]
        if self.tlm is not None:
            sci.tlm = self.tlm[mask]
            sci.tstamp = self.tstamp[mask]
            sci.images = tuple(self.images[ii] for ii in mask.nonzero()[0])
        return sci

    def vsel(self: SCItlm, mask: ArrayLike[bool]) -> SCItlm:
        """Return subset of SCItlm object generated using method vstack.

        Notes
        -----
        The methods `append` and `vsel` only work for complete measurements
        with the same MPS.

        """
        sci = SCItlm()
        if self.hdr is not None:
            sci.hdr = self.hdr[mask]
        if self.tlm is not None:
            sci.tlm = self.tlm[mask]
            sci.tstamp = self.tstamp[mask]
            sci.images = self.images[0][mask, :]
        return sci

    def append(self: SCItlm, sci: SCItlm) -> None:
        """Append one SCItlm object to the current.

        Notes
        -----
        The methods `append` and `vsel` only work for complete measurements
        with the same MPS.

        """
        self.hdr = sci.hdr if self.hdr is None else np.append(self.hdr, sci.hdr)
        self.tlm = sci.tlm if self.tlm is None else np.append(self.tlm, sci.tlm)
        self.tstamp = (
            sci.tstamp if self.tstamp is None else np.append(self.tstamp, sci.tstamp)
        )
        if isinstance(sci.images, tuple):
            self.images = (
                (sci.images[0],)
                if len(self.images) == 0
                else (np.concatenate((self.images[0], sci.images[0])),)
            )
        else:
            self.images = (
                sci.images
                if len(self.images) == 0
                else np.concatenate((self.images, sci.images))
            )

    def extract_l0_sci(
        self: SCItlm, ccsds_sci: tuple[NDArray], epoch: dt.datetime
    ) -> int:
        """Extract SPEXone level-0 Science-telemetry data.

        Parameters
        ----------
        ccsds_sci :  tuple[NDArray, ...]
           SPEXone level-0 Science-telemetry packets
        epoch :  dt.datetime
           Epoch of the telemetry packets (1958 or 1970)

        Returns
        -------
        int
            number of detector frames

        """
        self.init_attrs()
        if not ccsds_sci:
            return 0

        n_frames = 0
        hdr_dtype = None
        hk_dtype = None
        found_start_first = False
        for buf in ccsds_sci:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])
            if ccsds_hdr.grouping_flag == 1:
                found_start_first = True
                if n_frames == 0:
                    hdr_dtype = buf["hdr"].dtype
                    hk_dtype = buf["hk"].dtype
                    continue

            if not found_start_first:
                continue

            if ccsds_hdr.grouping_flag == 2:
                found_start_first = False
                n_frames += 1

        # do we have any complete detector images (Note ccsds_sci not empty!)?
        # print(f"n_frames: {n_frames}")
        if n_frames == 0:
            return 0

        # allocate memory
        self.hdr = np.empty(n_frames, dtype=hdr_dtype)
        self.tlm = np.empty(n_frames, dtype=hk_dtype)
        self.tstamp = np.empty(n_frames, dtype=TSTAMP_TYPE)

        # extract data from ccsds_sci
        ii = 0
        img = None
        found_start_first = False
        for buf in ccsds_sci:
            ccsds_hdr = CCSDShdr(buf["hdr"][0])
            if ccsds_hdr.grouping_flag == 1:
                found_start_first = True
                self.hdr[ii] = buf["hdr"]
                self.tlm[ii] = buf["hk"]
                self.tstamp[ii] = (
                    buf["icu_tm"]["tai_sec"][0],
                    buf["icu_tm"]["sub_sec"][0],
                    epoch
                    + dt.timedelta(
                        seconds=int(buf["icu_tm"]["tai_sec"][0]),
                        milliseconds=-self.readout_offset(ii),
                        microseconds=subsec2musec(buf["icu_tm"]["sub_sec"][0]),
                    ),
                )
                img = (buf["frame"][0],)
                continue

            if not found_start_first:
                continue

            if ccsds_hdr.grouping_flag == 0:
                img += (buf["frame"][0],)
            elif ccsds_hdr.grouping_flag == 2:
                found_start_first = False
                img += (buf["frame"][0],)
                img = np.concatenate(img)
                if img.size == self.tlm[ii]["IMRLEN"] // 2:
                    self.images += (img,)
                    ii += 1
                else:
                    n_frames -= 1
                if ii == n_frames:
                    break

        # adjust number of frames for corrupted images
        if ii != self.hdr.size:
            self.hdr = self.hdr[:ii]
            self.tlm = self.tlm[:ii]
            self.tstamp = self.tstamp[:ii]

        return n_frames

    def extract_l1a_sci(self: SCItlm, fid: h5py.File, mps_id: int | None) -> None:
        """Extract data from SPEXone level-1a Science-telemetry packets.

        Parameters
        ----------
        fid :  h5py.File
           A HDF5 file pointer to a SPEXone level-1a product
        mps_id : int, optional
           Select data performed with MPS equals 'mps_id'

        """
        # pylint: disable=no-member
        self.init_attrs()

        # read science telemetry
        dset = fid["/science_data/detector_telemetry"]
        if mps_id is None:
            data_sel = np.s_[:]
        else:
            data_sel = mask2slice(dset.fields("MPS_ID")[:] == mps_id)
            if data_sel is None:
                return
        self.tlm = fid["/science_data/detector_telemetry"][data_sel]

        # determine time-stamps
        dset = fid["/image_attributes/image_time"]
        indx = dset.attrs["units"].decode().find("20")
        ref_date = dt.datetime.fromisoformat(
            dset.attrs["units"].decode()[indx:] + "+00:00"
        )
        self.tstamp = np.empty(len(dset[data_sel]), dtype=TSTAMP_TYPE)
        self.tstamp["tai_sec"] = fid["/image_attributes/icu_time_sec"][data_sel]
        self.tstamp["sub_sec"] = fid["/image_attributes/icu_time_subsec"][data_sel]
        self.tstamp["dt"] = [
            ref_date + dt.timedelta(seconds=float(x)) for x in dset[data_sel]
        ]

        # read image data
        self.images = (fid["/science_data/detector_images"][data_sel, :],)

    def adc_gain(self: SCItlm, indx: int | None = None) -> NDArray[np.int32]:
        """Return ADC gain [Volt]."""
        if indx is None:
            indx = np.s_[:]
        return self.tlm["DET_ADCGAIN"][indx]

    def pga_gain(self: SCItlm, indx: int | None = None) -> NDArray[float]:
        """Return PGA gain [Volt]."""
        if indx is None:
            indx = np.s_[:]

        # need first bit of address 121
        reg_pgagainfactor = self.tlm["DET_BLACKCOL"][indx] & 0x1

        reg_pgagain = self.tlm["DET_PGAGAIN"][indx]

        return (1 + 0.2 * reg_pgagain) * 2**reg_pgagainfactor

    def exposure_time(self: SCItlm, indx: int | None = None) -> float | NDArray[float]:
        """Return exposure time [ms]."""
        if indx is None:
            indx = np.s_[:]
        return 129e-4 * (
            0.43 * self.tlm["DET_FOTLEN"][indx] + self.tlm["DET_EXPTIME"][indx]
        )

    def frame_period(self: SCItlm, indx: int | None = None) -> float | NDArray[float]:
        """Return frame period of detector measurement [ms]."""
        if indx is None:
            res = np.zeros(self.tlm.size, dtype="f8")
            _mm = (self.tlm["REG_FULL_FRAME"] & 0x3) == 2
            # binning mode
            if np.sum(_mm) > 0:
                res[_mm] = DET_CONSTS["FTI_science"]

            # full-frame mode
            _mm = ~_mm
            if np.sum(_mm) > 0:
                res[_mm] = np.clip(
                    DET_CONSTS["FTI_margin"]
                    + DET_CONSTS["FOT_diagnostic"]
                    + self.exposure_time(_mm),
                    a_min=DET_CONSTS["FTI_diagnostic"],
                    a_max=None,
                )
            return res

        # binning mode
        if (self.tlm["REG_FULL_FRAME"][indx] & 0x3) == 2:
            return DET_CONSTS["FTI_science"]

        # full-frame mode
        return np.clip(
            DET_CONSTS["FTI_margin"]
            + DET_CONSTS["FOT_diagnostic"]
            + self.exposure_time(indx),
            a_min=DET_CONSTS["FTI_diagnostic"],
            a_max=None,
        )

    def master_cycle(self: SCItlm, indx: int) -> float:
        """Return master-cycle period."""
        return self.tlm["REG_NCOADDFRAMES"][indx] * self.frame_period(indx)

    def readout_offset(self: SCItlm, indx: int) -> float:
        """Return offset wrt start-of-integration [ms]."""
        n_coad = self.tlm["REG_NCOADDFRAMES"][indx]
        n_frm = (
            n_coad + 1
            if self.tlm["IMRLEN"][indx] == FULLFRAME_BYTES
            else 2 * n_coad + 1
        )
        return n_frm * self.frame_period(indx)

    def binning_table(self: SCItlm) -> NDArray[np.int8]:
        """Return binning table identifier (zero for full-frame images)."""
        bin_tbl = np.zeros(len(self.tlm), dtype="i1")
        _mm = (self.tlm["IMRLEN"] == FULLFRAME_BYTES) | (self.tlm["IMRLEN"] == 0)
        if np.sum(_mm) == len(self.tlm):
            return bin_tbl

        bin_tbl_start = self.tlm["REG_BINNING_TABLE_START"]
        bin_tbl[~_mm] = 1 + (bin_tbl_start[~_mm] - 0x80000000) // 0x400000
        return bin_tbl

    def digital_offset(self: SCItlm) -> NDArray[np.int32]:
        """Return digital offset including ADC offset [count]."""
        buff = self.tlm["DET_OFFSET"].astype("i4")
        buff[buff >= 8192] -= 16384

        return buff + 70

    def convert(self: SCItlm, key: str) -> NDArray[float]:
        """Convert telemetry parameter to physical units.

        Parameters
        ----------
        key :  str
           Name of telemetry parameter

        Returns
        -------
        NDArray[float]

        """
        parm = key.upper()
        raw_data = np.array([x[parm] for x in self.tlm])
        return convert_hk(parm, raw_data)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Provides the common global-attributes for SPEXone Level-1 products."""

from __future__ import annotations

__all__ = ["attrs_def"]

import datetime as dt
import sys
from pathlib import Path

from . import pyspex_version


# - main functions --------------------------------
def attrs_def(inflight: bool = True, origin: str | None = None) -> dict:
    """Define global attributes of a SPEXone Level-1A product.

    Parameters
    ----------
    inflight : bool, default=True
       Flag for in-flight or on-ground products
    origin : str
       Product origin: 'SRON' or 'NASA'

    Returns
    -------
    dict
       Global attributes for a Level-1A product

    """
    if origin is None:
        origin = "NASA" if inflight else "SRON"

    res = {
        "publisher_name": "NASA/GSFC",
        "creator_name": "NASA/GSFC",
        "creator_email": "data@oceancolor.gsfc.nasa.gov",
        "creator_url": "https://oceancolor.gsfc.nasa.gov",
        "institution": (
            "NASA Goddard Space Flight Center, Ocean Biology Processing Group"
        ),
        "publisher_email": "data@oceancolor.gsfc.nasa.gov",
        "publisher_url": "https://oceancolor.gsfc.nasa.gov",
        "standard_name_vocabulary": "CF Standard Name Table v79",
        "keyword_vocabulary": (
            "NASA Global Change Master Directory (GCMD) Science Keywords"
        ),
        "date_created": dt.datetime.now(dt.UTC)
        .replace(tzinfo=None)
        .isoformat(timespec="milliseconds"),
        "license": (
            "https://www.earthdata.nasa.gov/engage/"
            "open-data-services-and-software/data-and-information-policy"
        ),
        "naming_authority": "gov.nasa.gsfc.sci.oceancolor",
        "project": "PACE Project",
        "conventions": "CF-1.10 ACDD-1.3",
        "title": "PACE SPEXone Level-1A Data",
        "platform": "PACE",
        "instrument": "SPEXone",
        "stdname_vocabulary": "NetCDF Climate and Forecast (CF) Metadata Convention",
        "processing_level": "L1A",
        "cdm_data_type": "swath" if inflight else "granule",
        "product_name": None,
        "start_direction": "Ascending" if inflight else None,
        "end_direction": "Ascending" if inflight else None,
        "time_coverage_start": "yyyy-mm-ddTHH:MM:DD",
        "time_coverage_end": "yyyy-mm-ddTHH:MM:DD",
        "history": " ".join(sys.argv),
        "processing_version": 1,
        "identifier_product_doi_authority": "https://dx.doi.org/",
        "identifier_product_doi": "10.5067/PACE/SPEXONE/L1A/SCI/2",
        # these will be writen as group attributes of /processing_control
        "software_name": f"{Path(sys.argv[0]).name}",
        "software_description": "SPEXone L0-L1A processor (SRON)",
        "software_url": "https://github.com/rmvanhees/pyspex",
        "software_version": pyspex_version(),
        "software_doi": "https://doi.org/10.5281/zenodo.5705691",
    }

    if origin == "SRON":
        res["institution"] = "SRON"
        res["creator_name"] = "SRON/Earth"
        res["creator_email"] = "SPEXone-MPC@sron.nl"
        res["creator_url"] = "https://www.sron.nl/missions-earth/pace-spexone"
        res["publisher_name"] = "SRON/Earth"
        res["publisher_email"] = "SPEXone-MPC@sron.nl"
        res["publisher_url"] = "https://www.sron.nl/missions-earth/pace-spexone"

    return res



# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Module with dedicated functions and classes to write a L1A product."""

from __future__ import annotations

__all__ = [
    "SpexL1A",
    "check_input_files",
    "create_l1a",
]

import datetime as dt
import logging
import sys
from dataclasses import asdict
from importlib.resources import files
from pathlib import Path
from typing import TYPE_CHECKING, Self

import h5py
import numpy as np
from h5yaml.conf_from_yaml import conf_from_yaml
from h5yaml.yaml_h5 import H5Yaml

from pyspex.lib import pyspex_version
from pyspex.tlm import get_l1a_filename

if TYPE_CHECKING:
    from dataclasses import dataclass

    from pyspex.lib.hk_tlm import HKtlm
    from pyspex.lib.sci_tlm import SCItlm
    from pyspex.tlm import SPXtlm

# - global parameters ------------------------------
module_logger = logging.getLogger("pyspex.l1a")


# - local functions --------------------------------
def check_input_files(config: dataclass) -> dataclass:
    """Check SPEXone level-0 files on existence and format.

    Parameters
    ----------
    config :  dataclass
       A dataclass that contains the settings of the L0-L1A processor

    Returns
    -------
    dataclass
       fields 'l0_format' {'raw', 'st3', 'dsb'} and 'l0_list' are updated.

    Raises
    ------
    FileNotFoundError
       If files are not found on the system.
    TypeError
       If determined file type differs from value supplied by user.

    """
    file_list = config.l0_list
    if file_list[0].suffix == ".H":
        if not file_list[0].is_file():
            raise FileNotFoundError(file_list[0])
        data_dir = file_list[0].parent
        file_stem = file_list[0].stem
        file_list = (
            sorted(data_dir.glob(file_stem + ".[0-9]"))
            + sorted(data_dir.glob(file_stem + ".?[0-9]"))
            + sorted(data_dir.glob(file_stem + "_hk.[0-9]"))
        )
        if not file_list:
            raise FileNotFoundError(file_stem + ".[0-9]")

        config.l0_format = "raw"
        config.l0_list = file_list
    elif file_list[0].suffix == ".ST3":
        if not file_list[0].is_file():
            raise FileNotFoundError(file_list[0])
        config.l0_format = "st3"
        config.l0_list = [file_list[0]]
    elif file_list[0].suffix == ".spx":
        file_list_out = []
        for flname in file_list:
            if not flname.is_file():
                raise FileNotFoundError(flname)

            if flname.suffix == ".spx":
                file_list_out.append(flname)

        if not file_list_out:
            raise FileNotFoundError(file_list)
        config.l0_format = "dsb"
        config.l0_list = file_list_out
    else:
        raise TypeError("Input files not recognized as SPEXone level-0 data")

    return config


def create_l1a(
    config: dataclass,
    tlm: SPXtlm,
    nav_dict: dict | None,
    mode: str | None = None,
) -> None:
    """All calls necessary to generate a SPEXone L1A product."""
    dims_nav = {}
    if config.hkt_list:
        dims_nav = {
            "/navigation_data/att_time": nav_dict["att_time"].size,
            "/navigation_data/orb_time": nav_dict["orb_time"].size,
            "/navigation_data/tilt_time": nav_dict["tilt_time"].size,
        }

    with SpexL1A(
        get_l1a_filename(config, tlm.coverage, mode),
        tlm.coverage,
        dims={
            "hk_packets": tlm.nomhk.size,
            "number_of_images": tlm.science.size,
            "samples_per_image": np.max([img.size for img in tlm.science.images]),
        }
        | dims_nav,
    ) as l1a:
        l1a.write_config(config)
        l1a.write_img_vars(tlm.science)  # before write_hk_vars
        l1a.write_hk_vars(tlm.nomhk)
        if config.hkt_list:
            l1a.write_nav_vars(nav_dict)


# - class SpexL1A ---------------------------------
# pylint: disable=no-member
class SpexL1A(H5Yaml):
    """Class to generate a SPEXone level-1A product.

    Parameters
    ----------
    l1a_name :  Path | str
       Name of the SPEXone level-1A product
    time_coverage :  list[dt.datetime, dt.datetime]
       Time coverage of the measurement data
    mode :  str, default="w"
       Support standard modes like 'r+' or 'w' (later is the default)
    dims :  dict[str, int], optional
       Change one or more unlimited dimensions to fixed-size dimensions
       Default of samples_per_image = row * column as defined in the YAML definition

    """

    def __init__(
        self: SpexL1A,
        l1a_name: Path | str,
        time_coverage: list[dt.datetime, dt.datetime],
        mode: str = "w",
        *,
        dims: dict[str, int] | None = None,
    ) -> None:
        """Initialize SpexL1A object and create empty SPEXone level-1A product."""
        self.logger = logging.getLogger("pyspex.SpexL1A")
        self.filename = l1a_name if isinstance(l1a_name, Path) else Path(l1a_name)

        super().__init__(files("pyspex.Data") / "h5_level_1a.yaml")

        # convert unlimited dimensions to fixed-size dimensions
        if "samples_per_image" not in dims:
            dims["samples_per_image"] = (
                self.h5_def["dimensions"]["column"]["_size"]
                * self.h5_def["dimensions"]["row"]["_size"]
            )
        for key, value in dims.items():
            if self.h5_def["dimensions"][key]["_size"] == 0:
                self.h5_def["dimensions"][key]["_size"] = value

        # create empty file and (re-)open file in append mode
        if mode == "w":
            self.create(self.filename)
        self.fid = h5py.File(self.filename, "r+")
        self.fid.attrs["time_coverage_start"] = (
            time_coverage[0].replace(tzinfo=None).isoformat(timespec="milliseconds")
        )
        self.fid.attrs["time_coverage_end"] = (
            time_coverage[1].replace(tzinfo=None).isoformat(timespec="milliseconds")
        )

    def __enter__(self: SpexL1A) -> Self:
        """Initiate the context manager."""
        return self

    def __exit__(self: SpexL1A, *args: object) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: SpexL1A) -> None:
        """Close resources, after sanity check of L1A product."""
        self.add_attrs()
        self.fid.close()

    def add_attrs(self: SpexL1A) -> None:
        """Add global attributes to HDF5 product."""
        try:
            attrs_def = conf_from_yaml(files("pyspex.Data") / "h5_global_attrs.yaml")
        except RuntimeError as exc:
            raise RuntimeError from exc

        # variable globale attributes
        self.fid.attrs["product_name"] = self.filename.name
        self.fid.attrs["date_created"] = (
            dt.datetime.now(dt.UTC)
            .replace(tzinfo=None)
            .isoformat(timespec="milliseconds")
        )
        # constant global attributes
        for key, value in attrs_def["attrs_global"].items():
            if key not in self.fid.attrs:
                self.fid.attrs[key] = value
        for key, value in attrs_def["attrs_nasa"].items():
            self.fid.attrs[key] = value
        self.fid.attrs["history"] = " ".join(sys.argv)

        gid = self.fid["/processing_control"]
        gid.attrs["software_name"] = Path(sys.argv[0]).name
        gid.attrs["software_version"] = pyspex_version()
        for key, value in attrs_def["attrs_cntrl"].items():
            gid.attrs[key] = value

    def write_config(self: SpexL1A, config: dataclass) -> None:
        """Write command-line settings to /processing_control/input_parameters."""
        self.fid.attrs["processing_version"] = config.processing_version
        gid = self.fid["/processing_control/input_parameters"]
        for key, value in asdict(config).items():
            if key in ("l0_list", "hkt_list"):
                gid.attrs[key] = "" if not value else ",".join([x.name for x in value])
            else:
                gid.attrs[key] = f"{value}"

    def write_hk_vars(self: SpexL1A, nomhk: HKtlm) -> None:
        """Write nominal housekeeping data.

        Parameters
        ----------
        nomhk :  HKtlm
           SPEXone nominal housekeeping telemetry records

        """
        group = "/engineering_data"

        hk_time = np.array(
            [x.replace(tzinfo=None) for x in nomhk.tstamp], dtype="datetime64[us]"
        )

        # obtain last midnight and update the units accordingly
        dset = self.fid[f"{group}/HK_tlm_time"]
        if dset.attrs["units"].find("%Y-%m-%d %H:%M:%S") > 0:
            midnight = hk_time[0].astype("datetime64[D]")
            dset.attrs["units"] = dset.attrs["units"].replace(
                "%Y-%m-%d %H:%M:%S", f"{midnight:s} 00:00:00"
            )
        else:
            midnight = np.datetime64(dset.attrs["units"].split(" ")[2])
        dset[:] = (hk_time - midnight).astype(float) / 1e6

        self.fid[f"{group}/NomHK_telemetry"][:] = nomhk.tlm
        self.fid[f"{group}/temp_detector"][:] = nomhk.convert("TS1_DEM_N_T")
        self.fid[f"{group}/temp_housing"][:] = nomhk.convert("TS2_HOUSING_N_T")
        self.fid[f"{group}/temp_radiator"][:] = nomhk.convert("TS3_RADIATOR_N_T")
        self.logger.debug("wrote data to group: %s.", group)

    def write_img_vars(self: SpexL1A, science: SCItlm) -> None:
        """Write Science image and Science housekeeping data.

        Parameters
        ----------
        science :  SCItlm
           SPEXone Science telemetry records

        """
        group = "/image_attributes"

        image_time = np.array(
            [x.replace(tzinfo=None) for x in science.tstamp["dt"]],
            dtype="datetime64[us]",
        )

        # obtain last midnight and update the units accordingly
        dset = self.fid[f"{group}/image_time"]
        if dset.attrs["units"].find("%Y-%m-%d %H:%M:%S") > 0:
            midnight = image_time[0].astype("datetime64[D]")
            dset.attrs["units"] = dset.attrs["units"].replace(
                "%Y-%m-%d %H:%M:%S", f"{midnight:s} 00:00:00"
            )
        else:
            midnight = np.datetime64(dset.attrs["units"].split(" ")[2])
        dset[:] = (image_time - midnight).astype(float) / 1e6

        self.fid[f"{group}/timedelta_centre"][:] = (
            (science.tlm["REG_NCOADDFRAMES"] - 1) * science.frame_period()
            + science.exposure_time()
        ) / 2000
        self.fid[f"{group}/icu_time_sec"][:] = science.tstamp["tai_sec"]
        self.fid[f"{group}/icu_time_subsec"][:] = science.tstamp["sub_sec"]
        self.fid[f"{group}/image_id"][:] = np.bitwise_and(
            science.hdr["sequence"], 0x3FFF
        )
        self.fid[f"{group}/binning_table"][:] = science.binning_table()
        self.fid[f"{group}/digital_offset"][:] = science.digital_offset()
        self.fid[f"{group}/nr_coadditions"][:] = science.tlm["REG_NCOADDFRAMES"]
        self.fid[f"{group}/exposure_time"][:] = science.exposure_time() / 1000
        self.logger.debug("wrote data to group: %s.", group)

        group = "/science_data"
        self.fid[f"{group}/science_hk"][:] = science.tlm
        if len(np.unique([img.size for img in science.images])) == 1:
            images = science.images
        else:
            nj_max = 0
            for img in science.images:
                nj_max = max(nj_max, img.size)
            images = np.full((science.size, nj_max), np.iinfo("u2").max)
            for ii, img in enumerate(science.images):
                images[ii, : img.size] = img
        self.fid[f"{group}/detector_images"][:] = images
        self.logger.debug("wrote data to group: %s.", group)

    def write_nav_vars(
        self: SpexL1A,
        nav_dict: dict,
    ) -> None:
        """Write navigation-data.

        Parameters
        ----------
        nav_dict :  dict[str, NDArray]
           Dictionary holding all navigation parameters

        """
        group = "/navigation_data"

        gid = self.fid[group]
        for key, value in nav_dict.items():
            dset = gid[key]
            if key.endswith("_time"):
                if dset.attrs["units"].find("%Y-%m-%d %H:%M:%S") > 0:
                    midnight = value[0].astype("datetime64[D]")
                    dset.attrs["units"] = dset.attrs["units"].replace(
                        "%Y-%m-%d %H:%M:%S", f"{midnight:s} 00:00:00"
                    )
                else:
                    midnight = np.datetime64(dset.attrs["units"].split(" ")[2])
                dset[:] = (value - midnight).astype(float) / 1e6
            elif key == "coverage_quality":
                dset[()] = value
            else:
                dset[:] = value

        # write or update attitude coverage-time
        midnight = np.datetime64(gid["att_time"].attrs["units"].split(" ")[2])
        gid.attrs["time_coverage_start"] = str(
            midnight + (1e3 * gid["att_time"][0]).astype("timedelta64[ms]")
        )
        gid.attrs["time_coverage_end"] = str(
            midnight + (1e3 * gid["att_time"][-1]).astype("timedelta64[ms]")
        )
        self.logger.debug("wrote data to group: %s.", group)



#
# This file is part of pyspex:
#    https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""Generate a SPEXone level-1A product (netCDF4 format)."""

from __future__ import annotations

import logging
import warnings

import numpy as np

from pyspex.hkt_io import HKTio
from pyspex.lib.logger import start_logger
from pyspex.lv0_lib import CorruptPacketWarning
from pyspex.tlm import SPXtlm

from .argparse_gen_l1a import argparse_gen_l1a
from .l1a import check_input_files, create_l1a


# - main function ----------------------------------
def main() -> int:
    """Execute the main bit of the application."""
    error_code = 0
    warn_code = 0

    # (1) initialize logger
    start_logger()
    logging.captureWarnings(True)

    # (2) parse command-line parameters and YAML file for settings
    config = argparse_gen_l1a()
    logging.getLogger().setLevel(config.verbose)  # first, set the root logger
    logger = logging.getLogger("pyspex.gen_l1a")  # then initiate a descendant
    logger.debug("%s", config)

    # (3) check input files (SEPXone level-0)
    try:
        check_input_files(config)
    except FileNotFoundError as exc:
        logger.fatal('File "%s" not found on system.', exc)
        return 110
    except TypeError as exc:
        logger.fatal("%s", exc)
        return 121

    # (4) read level 0 data
    tlm = None
    with warnings.catch_warnings(record=True) as wrec_list:
        warnings.simplefilter("always", category=CorruptPacketWarning)
        try:
            tlm = SPXtlm()
            tlm.from_lv0(
                config.l0_list,
                file_format=config.l0_format,
                debug=config.debug,
                dump=config.dump,
            )
        except FileNotFoundError as exc:
            logger.fatal('FileNotFoundError exception raised for "%s".', exc)
            error_code = 110
        except TypeError as exc:
            logger.fatal('TypeError exception raised with "%s".', exc)
            error_code = 121

        for wrec in wrec_list:
            logger.warning('CorruptPacketWarning raised with "%s".', str(wrec.message))
            warn_code = 122

    if error_code != 0 or config.debug or config.dump:
        return error_code

    if tlm.science.size > 0:
        mps_list = np.unique(tlm.science.tlm["MPS_ID"])
        logger.debug("unique Science MPS: %s", mps_list)
        tlm.nomhk = tlm.nomhk.sel(np.isin(tlm.nomhk.tlm["MPS_ID"], mps_list))

    # (5) read navigation data from PACE_HKT products
    nav_dict = None
    if config.hkt_list:
        coverage_spx = (
            tlm.coverage[0].replace(tzinfo=None),
            tlm.coverage[1].replace(tzinfo=None),
        )
        hkt = HKTio(config.hkt_list)
        nav_dict = hkt.navigation()
        nav_dict = hkt.nav_coverage_adjust(nav_dict, coverage_spx)
        nav_dict["coverage_quality"] = hkt.nav_coverage_flag(coverage_spx)

    # (6) write Level-1A product.
    try:
        if config.eclipse is None or tlm.science.size == 0:
            create_l1a(config, tlm, nav_dict)
        elif config.eclipse:
            # binned measurements
            create_l1a(config, tlm.binned(), nav_dict, "binned")

            # full-frame measurements
            create_l1a(config, tlm.full(), nav_dict, "full")
        else:
            # binned measurements
            create_l1a(config, tlm.binned(), nav_dict, "binned")
    except (KeyError, RuntimeError) as exc:
        # raise RuntimeError from exc
        logger.fatal('RuntimeError with "%s"', exc)
        error_code = 131
    except UserWarning as exc:
        logger.warning('navigation data is incomplete: "%s".', exc)
        error_code = 132
    except Exception as exc:
        # raise RuntimeError from exc
        logger.fatal('Unexpected exception occurred with "%s".', exc)
        error_code = 135

    return warn_code if error_code == 0 else error_code



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains helper functions for the class `SPXtlm`."""

from __future__ import annotations

__all__ = ["CONV_DICT", "HkFlagging", "convert_hk"]

from enum import Enum, IntFlag, auto
from typing import TYPE_CHECKING

import numpy as np
from numpy import ma

if TYPE_CHECKING:
    from collections.abc import Callable


# - helper functions ------------------------
def convert_2_float(raw_data: np.ndarray) -> np.ndarray:
    """Convert integer array to float."""
    return raw_data.astype(float)


def exp_spex_det_t(raw_data: np.ndarray) -> np.ndarray:
    """Convert Detector Temperature Sensor to degree Celsius."""
    res = np.empty(raw_data.size, dtype=float)
    mask = raw_data < 400
    res[mask] = 1.224 * raw_data[mask] - 290.2
    res[~mask] = 0.6426 * raw_data[~mask] - 418.72
    res[res > 50] = np.nan
    return res


def exp_spex_thermistor(raw_data: np.ndarray) -> np.ndarray:
    """Convert readouts of the Temperature Sensors to degree Celsius.

    Notes
    -----
    - TS1 DEM Nominal temperature
    - TS2 Housing Nominal Temperature
    - TS3 Radiator Nominal Temperature
    - TS4 DEM Redundant Temperature
    - TS5 Housing Redundant Temperature
    - TS6 Radiator Redundant Temperature*

    """
    coefficients = (21.19, 272589.0, 1.5173e-15, 5.73666e-19, 5.11328e-20)
    buff = ma.masked_array(raw_data / 256, mask=raw_data == 0)
    buff = (
        coefficients[0]
        + coefficients[1] / buff
        - coefficients[2] * buff**4
        + (coefficients[3] - coefficients[4] * ma.log(buff)) * buff**5
    )
    buff[raw_data == 0] = np.nan
    return buff.data


def poly_spex_icuhk_internaltemp(raw_data: np.ndarray) -> np.ndarray:
    """Convert readouts of temperature sensors on ICU power supplies to degree Celsius.

    Notes
    -----
    - ICU V5 supply temperature
    - ICU V4 supply temperature
    - ICU HtrG1 supply temperature
    - ICU HtrG2 supply temperature
    - ICU MidBoard temperature
    - ICU MCU-RAM temperature
    - ICU 1V2, 3V3 supply temperature

    """
    coefficients = (0.0, 0.0625)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_icuhk_internaltemp2(raw_data: np.ndarray) -> np.ndarray:
    """Convert readings of ICU bus voltages to Voltages.

    Notes
    -----
    - ICU 4.0 Volt bus voltage
    - ICU 3.3 Volt bus voltage
    - ICU 1.2 Volt bus voltage
    - DEM power supply

    """
    coefficients = (0, 0.001)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_htr_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater Bus voltages to Volt."""
    coefficients = (0, 0.01 / 3)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_htr_i(raw_data: np.ndarray) -> np.ndarray:
    """Convert readings of Heater Currents to mA."""
    coefficients = (0, 0.001)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_htr1_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater1 Current to Watt."""
    resistance = 238.3
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr2_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater2 Current to Watt."""
    resistance = 212.5
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr3_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater3 Current to Watt."""
    resistance = 237.7
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr4_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater4 Current to Watt."""
    resistance = 213.2
    return resistance * (raw_data / 1000) ** 2


def poly_spex_dutycycle(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater Controller Duty Cycle output to percent."""
    coefficients = (0, 0.1)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_anode_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED Anode voltage to Volt."""
    coefficients = (0, 0.000000623703003)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_cath_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED Cathode voltage to Volt."""
    coefficients = (0, 0.000000415802001953)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_i(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED current to milli-Ampere."""
    coefficients = (0, 0.0000030307446495961334745762711864407)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_vcc(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC Analog VCC reading to Volt."""
    coefficients = (0, 0.00000127156575520833333)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_gain(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC Analog VCC reading to Volt."""
    coefficients = (0, 0.000000127156575520833333)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_t(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC1 Temperature reading to degree Celsius."""
    coefficients = (-273.4, 0.0007385466842651367)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_offset(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC offset (?) to Voltage."""
    coefficients = (0, 0.000415802001953)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_dem_i(raw_data: np.ndarray) -> np.ndarray:
    """Convert DEM Supply current to milli-Ampere."""
    coefficients = (0, 0.2417)
    return coefficients[0] + coefficients[1] * raw_data


class Enable(Enum):
    """Possible values of the DEM SpaceWire link status."""

    INACTIVE = 0
    ACTIVE = 1


class EnaDis(Enum):
    """Possible values of the LED commanded enable state."""

    OFF = 0
    ON = 1


class IcuMode(Enum):
    """Possible values of the ICU mode."""

    IDLE = 0
    OPER = 1


class IsEna(Enum):
    """Possible values of the Power Sense."""

    DISABLED = 0
    ENABLED = 1


class OcdReg(Enum):
    """Possible values of the OCD state."""

    PWR_OK = 0
    OC2V1 = 1
    OC3V3 = 2
    OC2V1_AND_3V3 = 3


class TimeMsg(Enum):
    """Received status of external PPS or Time message."""

    MISSING = 0
    RECEIVED = 1


class WriteProt(Enum):
    """Status of Flash Write protection (both banks)."""

    BANK0_DISABLED = 0
    BANK0_ENABLED = 1
    BANK1_DISABLED = 16
    BANK1_ENABLED = 17


# - exported functions ----------------------
CONV_DICT: dict[
    str, dict[str | None, Callable[..., np.ndarray] | None, tuple[int] | None]
] = {
    "SEQCNT": {"units": None, "func": None, "range": None},
    "TCPKTID": {"units": None, "func": None, "range": None},
    "TCPKTSEQCTRL": {"units": None, "func": None, "range": None},
    "TCREJCODE": {"units": None, "func": None, "range": None},
    "TCFAILCODE": {"units": None, "func": None, "range": None},
    "TCREJPKTID": {"units": None, "func": None, "range": None},
    "TCFAILPKTID": {"units": None, "func": None, "range": None},
    "TCACCCNT": {"units": None, "func": None, "range": None},
    "TCREJCNT": {"units": None, "func": None, "range": None},
    "TCEXECCNT": {"units": None, "func": None, "range": None},
    "TCFAILCNT": {"units": None, "func": None, "range": None},
    "ICUSWVER": {"units": None, "func": None, "range": None},
    "SYSSTATE": {"units": None, "func": None, "range": None},
    "ICUMODE": {"units": None, "func": None, "range": None},
    "EXTPPSSTAT": {"units": None, "func": None, "range": None},
    "TIMEMSGSTAT": {"units": None, "func": None, "range": None},
    "OBTSYNCSTAT": {"units": None, "func": None, "range": None},
    "MPS_ID": {"units": None, "func": None, "range": None},
    "MPS_VER": {"units": None, "func": None, "range": None},
    "EVNTCNT_DEBUG": {"units": None, "func": None, "range": None},
    "EVNTCNT_PROG": {"units": None, "func": None, "range": None},
    "EVNTCNT_WARN": {"units": None, "func": None, "range": None},
    "EVNTCNT_ERR": {"units": None, "func": None, "range": None},
    "EVNTCNT_FATAL": {"units": None, "func": None, "range": None},
    "BOOTSTATEPREV": {"units": None, "func": None, "range": None},
    "BOOTCNTGOOD_IM0": {"units": None, "func": None, "range": None},
    "BOOTCNTGOOD_IM1": {"units": None, "func": None, "range": None},
    "BOOTCNTGOOD_IM2": {"units": None, "func": None, "range": None},
    "BOOTCNTGOOD_IM3": {"units": None, "func": None, "range": None},
    "BOOTATTEMPTS_CURRIM": {"units": None, "func": None, "range": None},
    "MPS_ACT_STATUS": {"units": None, "func": None, "range": None},
    "SWIMG_LOADED": {"units": None, "func": None, "range": None},
    "SWIMG_DEFAULT": {"units": None, "func": None, "range": None},
    "SWIMG_NXTBOOT": {"units": None, "func": None, "range": None},
    "WRITEPROT": {"units": None, "func": None, "range": None},
    "BOOTCAUSE": {"units": None, "func": None, "range": None},
    "TCVER_STAT": {"units": None, "func": None, "range": None},
    "SPW_REG_A": {"units": None, "func": None, "range": None},
    "SPW_REG_B": {"units": None, "func": None, "range": None},
    "LAST_CRC": {"units": None, "func": None, "range": None},
    "SCITM_PKTINTVL": {"units": None, "func": None, "range": None},
    "SCITM_BUFFREE": {"units": None, "func": None, "range": None},
    "SWEXECTIMEWC": {"units": None, "func": None, "range": None},
    "ERRCNT_PLACEHOLDER_03": {"units": None, "func": None, "range": None},
    "TS1_DEM_N_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (18.13, 18.53),
    },
    "TS2_HOUSING_N_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (19.51, 19.71),
    },
    "TS3_RADIATOR_N_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (-1.5, 2.0),
    },
    "TS4_DEM_R_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (17.7, 18.1),
    },
    "TS5_HOUSING_R_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (19.05, 19.55),
    },
    "TS6_RADIATOR_R_T": {
        "units": "degC",
        "func": exp_spex_thermistor,
        "range": (-1.5, 2.0),
    },
    "ICU_5V_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_4V_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_HG1_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_HG2_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_MID_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_MCU_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_DIGV_T": {
        "units": "degC",
        "func": poly_spex_icuhk_internaltemp,
        "range": (-20, 60),
    },
    "ICU_4P0V_V": {
        "units": "Volt",
        "func": poly_spex_icuhk_internaltemp2,
        "range": (3.86, 5.4),
    },
    "ICU_3P3V_V": {
        "units": "Volt",
        "func": poly_spex_icuhk_internaltemp2,
        "range": (3.1, 3.456),
    },
    "ICU_1P2V_V": {
        "units": "Volt",
        "func": poly_spex_icuhk_internaltemp2,
        "range": (1.21, 1.26),
    },
    "ICU_4P0V_I": {"units": "mA", "func": convert_2_float, "range": (100, 350)},
    "ICU_3P3V_I": {"units": "mA", "func": convert_2_float, "range": (100, 300)},
    "ICU_1P2V_I": {"units": "mA", "func": convert_2_float, "range": (70, 200)},
    "DEM_STATUS": {"units": None, "func": None, "range": None},
    "ICU_5P0V_V": {
        "units": "Volt",
        "func": poly_spex_icuhk_internaltemp2,
        "range": (4.85, 5.25),
    },
    "ICU_5P0V_I": {"units": "mA", "func": convert_2_float, "range": (5, 570)},
    "DEMSPWSTAT": {"units": None, "func": None, "range": None},
    "DEMRESETCNT": {"units": None, "func": None, "range": None},
    "HTRGRP1_V": {"units": "Volt", "func": poly_spex_htr_v, "range": (46, 49)},
    "HTRGRP2_V": {"units": "Volt", "func": poly_spex_htr_v, "range": (46, 49)},
    "HTR1_I": {"units": "mA", "func": poly_spex_htr_i, "range": (0, 250)},
    "HTR2_I": {"units": "mA", "func": poly_spex_htr_i, "range": (0, 250)},
    "HTR3_I": {"units": "mA", "func": poly_spex_htr_i, "range": (0, 250)},
    "HTR4_I": {"units": "mA", "func": poly_spex_htr_i, "range": (0, 250)},
    "HTR1_POWER": {"units": "Watt", "func": poly_spex_htr1_p, "range": (-1, 1)},
    "HTR2_POWER": {"units": "Watt", "func": poly_spex_htr2_p, "range": (0, 10)},
    "HTR3_POWER": {"units": "Watt", "func": poly_spex_htr3_p, "range": (0, 10)},
    "HTR4_POWER": {"units": "Watt", "func": poly_spex_htr4_p, "range": (-1, 1)},
    "HTR1_CALCPVAL": {"units": None, "func": None, "range": None},
    "HTR2_CALCPVAL": {"units": None, "func": None, "range": None},
    "HTR3_CALCPVAL": {"units": None, "func": None, "range": None},
    "HTR4_CALCPVAL": {"units": None, "func": None, "range": None},
    "HTR1_CALCIVAL": {"units": None, "func": None, "range": None},
    "HTR2_CALCIVAL": {"units": None, "func": None, "range": None},
    "HTR3_CALCIVAL": {"units": None, "func": None, "range": None},
    "HTR4_CALCIVAL": {"units": None, "func": None, "range": None},
    "HTR1_DUTYCYCL": {"units": "%", "func": poly_spex_dutycycle, "range": (-1, 1)},
    "HTR2_DUTYCYCL": {"units": "%", "func": poly_spex_dutycycle, "range": (6, 12)},
    "HTR3_DUTYCYCL": {"units": "%", "func": poly_spex_dutycycle, "range": (6, 12)},
    "HTR4_DUTYCYCL": {"units": "%", "func": poly_spex_dutycycle, "range": (-1, 1)},
    "LED1_ENADIS": {"units": None, "func": None, "range": None},
    "LED2_ENADIS": {"units": None, "func": None, "range": None},
    "LED1_ANODE_V": {
        "units": "Volt",
        "func": poly_spex_led_anode_v,
        "range": (2.9, 5.4),
    },
    "LED1_CATH_V": {"units": "Volt", "func": poly_spex_led_cath_v, "range": (0.7, 5.4)},
    "LED1_I": {"units": "mA", "func": poly_spex_led_i, "range": (0, 10.1)},
    "LED2_ANODE_V": {
        "units": "Volt",
        "func": poly_spex_led_anode_v,
        "range": (2.9, 5.4),
    },
    "LED2_CATH_V": {"units": "Volt", "func": poly_spex_led_cath_v, "range": (0.7, 5.4)},
    "LED2_I": {"units": "mA", "func": poly_spex_led_i, "range": (0, 10.1)},
    "ADC1_VCC": {"units": "Volt", "func": poly_spex_adc_vcc, "range": (4.77, 5.25)},
    "ADC1_GAIN": {"units": "Volt", "func": poly_spex_adc_gain, "range": None},
    "ADC1_REF": {"units": "Volt", "func": poly_spex_adc_vcc, "range": None},
    "ADC1_T": {"units": "degC", "func": poly_spex_adc_t, "range": (-20, 60)},
    "ADC1_OFFSET": {"units": "Volt", "func": poly_spex_adc_offset, "range": None},
    "ADC2_VCC": {"units": "Volt", "func": poly_spex_adc_vcc, "range": None},
    "ADC2_GAIN": {"units": "Volt", "func": poly_spex_adc_gain, "range": None},
    "ADC2_REF": {"units": "Volt", "func": poly_spex_adc_vcc, "range": (4.77, 5.25)},
    "ADC2_T": {"units": "degC", "func": poly_spex_adc_t, "range": (-20, 60)},
    "ADC2_OFFSET": {"units": "Volt", "func": poly_spex_adc_offset, "range": None},
    "DEM_V": {
        "units": "Volt",
        "func": poly_spex_icuhk_internaltemp2,
        "range": (4.7, 5.25),
    },
    "DEM_I": {"units": "mA", "func": poly_spex_dem_i, "range": (100, 550)},
    "REG_FW_VERSION": {"units": None, "func": None, "range": None},
    "DET_T": {"units": "degC", "func": exp_spex_det_t, "range": (19, 36)},
    "REG_SPW_ERROR": {"units": None, "func": None, "range": None},
    "REG_CMV_OUTOFSYNC": {"units": None, "func": None, "range": None},
    "REG_OCD_ACTUAL": {"units": None, "func": None, "range": None},
    "REG_OCD_STICKY": {"units": None, "func": None, "range": None},
    "REG_PWR_SENS": {"units": None, "func": None, "range": None},
    "REG_FLASH_STATUS": {"units": None, "func": None, "range": None},
    "REG_FLASH_EDAC_BLOCK": {"units": None, "func": None, "range": None},
    "SW_MAIN_LOOP_COUNT": {"units": None, "func": None, "range": None},
}


def convert_hk(parm: str, raw_data: np.ndarray) -> np.ndarray:
    """Convert a NomHK (or DemHK subset) parameter to physical units."""
    if (res := CONV_DICT.get(parm)) is None:
        raise KeyError(f"Parameter: {parm} not found in CONV_DICT")

    if (func := res["func"]) is not None:
        return func(raw_data)

    return raw_data


class HkFlagging(IntFlag):
    """..."""

    NOMINAL = 0
    CURRENT_TOO_LOW = auto()
    CURRENT_TOO_HIGH = auto()
    TEMP_TOO_LOW = auto()
    TEMP_TOO_HIGH = auto()
    VOLT_TOO_LOW = auto()
    VOLT_TOO_HIGH = auto()
    WATT_TOO_LOW = auto()
    WATT_TOO_HIGH = auto()
    CHANGED = auto()

    @classmethod
    def get_flag(cls: HkFlagging, parm: str, too_low: bool = False) -> HkFlagging:
        """Get value of flag for parameter out-of-range or changed parameter-values."""
        if (res := CONV_DICT.get(parm)) is None:
            raise KeyError(f"Parameter: {parm} not found in CONV_DICT")

        match res["units"]:
            case "mA" | "A":
                value = cls.CURRENT_TOO_LOW if too_low else cls.CURRENT_TOO_HIGH
            case "degC" | "K":
                value = cls.TEMP_TOO_LOW if too_low else cls.TEMP_TOO_HIGH
            case "Volt":
                value = cls.VOLT_TOO_LOW if too_low else cls.VOLT_TOO_HIGH
            case "Watt":
                value = cls.WATT_TOO_LOW if too_low else cls.WATT_TOO_HIGH
            case _:
                value = cls.CHANGED

        return value



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `HKTio` to read PACE HKT products."""

from __future__ import annotations

__all__ = ["CoverageFlag", "HKTio"]

import logging
from enum import IntFlag, auto
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np

from .lib.ccsds_hdr import CCSDShdr

if TYPE_CHECKING:
    from numpy.typing import NDArray

# - global parameters -----------------------
module_logger = logging.getLogger("pyspex.hkt_io")

# reduce extend of navigation data to
TEN_SECONDS = np.timedelta64(10, "s")
# require an extent of navigation data of
SEVEN_SECONDS = np.timedelta64(7, "s")


# - class CoverageFlag ----------------------
class CoverageFlag(IntFlag):
    """Define flags for coverage_quality (navigation_data)."""

    MISSING_SAMPLES = auto()
    TOO_SHORT_EXTENDS = auto()
    NO_EXTEND_AT_START = auto()
    NO_EXTEND_AT_END = auto()

    @classmethod
    def check(
        cls: CoverageFlag,
        coverage_nav: tuple[np.datetime64, np.datetime64],
        coverage_spx: tuple[np.datetime64, np.datetime64],
    ) -> int:
        """Check coverage time of navigation data.

        Parameters
        ----------
        coverage_nav: tuple[np.datetime64, np.datetime64]
           time-coverage of the PACE-HKT data
        coverage_spx: tuple[np.datetime64, np.datetime64]
           time-coverage of the SPEXone measurement data

        """
        coverage_quality = cls(0)

        # check at the start of the data
        if coverage_spx[0] < coverage_nav[0]:
            coverage_quality |= CoverageFlag.NO_EXTEND_AT_START
            module_logger.error(
                "time coverage of navigation data starts after L1A science data"
            )

        diff_coverage = coverage_spx[0] - coverage_nav[0]
        if diff_coverage <= SEVEN_SECONDS:
            coverage_quality |= CoverageFlag.TOO_SHORT_EXTENDS
            module_logger.warning(
                "extends of time coverage of navigation data too short",
            )

        # check at the end of the data
        if coverage_spx[1] > coverage_nav[1]:
            coverage_quality |= CoverageFlag.NO_EXTEND_AT_END
            module_logger.error(
                "time coverage of navigation data ends before L1A science data"
            )

        diff_coverage = coverage_nav[1] - coverage_spx[1]
        if diff_coverage <= SEVEN_SECONDS:
            coverage_quality |= CoverageFlag.TOO_SHORT_EXTENDS
            module_logger.warning(
                "extends of time coverage of navigation data too short",
            )
        return coverage_quality


# - class HKTio -------------------------
class HKTio:
    """Class to read housekeeping and navigation data from PACE-HKT products.

    Parameters
    ----------
    flnames :  str | Path | list[Path]
        list of PACE_HKT filenames (netCDF4 format)

    Notes
    -----
    This class has the following methods::

     - housekeeping(instrument: str) -> tuple[np.ndarray]
     - navigation() -> dict

    """

    def __init__(self: HKTio, flnames: str | Path | list[Path]) -> None:
        """Initialize access to PACE HKT products."""
        self.time_coverage = ()
        if isinstance(flnames, str | Path):
            self.flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        else:
            self.flnames = flnames

        for hkt_fl in self.flnames:
            if not hkt_fl.is_file():
                raise FileNotFoundError(f"file {hkt_fl} not found on system")

    @staticmethod
    def __read_hkt__(hkt_file: Path, instrument: str) -> np.ndarray | None:
        """Return housekeeping data of a given instrument."""
        with h5py.File(hkt_file) as fid:
            gid = fid["housekeeping_data"]
            if instrument == "spx":
                ds_name = (
                    "SPEX_HKT_packets"
                    if "SPEX_HKT_packets" in gid
                    else "SPEXone_HKT_packets"
                )
            elif instrument == "sc":
                ds_name = "SC_HKT_packets"
            elif instrument == "oci":
                ds_name = "OCI_HKT_packets"
            elif instrument == "harp":
                ds_name = (
                    "HARP_HKT_packets"
                    if "HARP_HKT_packets" in gid
                    else "HARP2_HKT_packets"
                )
            else:
                raise KeyError("data of unknown instrument requested")

            res = gid[ds_name][:] if ds_name in gid else None

        return res

    # ---------- PUBLIC FUNCTIONS ----------
    def housekeeping(self: HKTio, instrument: str = "spx") -> tuple[np.ndarray]:
        """Read housekeeping telemetry data from the HKT products.

        Parameters
        ----------
        instrument : {'spx', 'oci', 'harp', 'sc'}, default='spx'
           name of PACE instrument: 'harp': HARP2, 'oci': OCI,
           'sc': spacecraft, 'spx': SPEXone.

        Notes
        -----
        Current implementation only works for SPEXone.

        """
        res = None
        for hkt_file in self.flnames:
            buff = self.__read_hkt__(hkt_file, instrument)
            if buff is not None:
                res = buff if res is None else np.concatenate((res, buff), axis=0)
        if res is None:
            return ()

        ccsds_hk = ()
        for packet in res:
            try:
                ccsds_hdr = CCSDShdr()
                ccsds_hdr.read("raw", packet)
            except ValueError as exc:
                module_logger.warning('CCSDS header read error with "%s"', exc)
                break

            try:
                dtype_apid = ccsds_hdr.data_dtype
            except ValueError:
                print(
                    f"APID: 0x{ccsds_hdr.apid:x};"
                    f" Packet Length: {ccsds_hdr.packet_size:d}"
                )
                dtype_apid = None

            if dtype_apid is not None:  # all valid APIDs
                buff = np.frombuffer(packet, count=1, offset=0, dtype=dtype_apid)
                ccsds_hk += (buff,)
            else:
                module_logger.warning(
                    "package with APID 0x%x and length %d is not implemented",
                    ccsds_hdr.apid,
                    ccsds_hdr.packet_size,
                )

        return ccsds_hk

    def navigation(self: HKTio) -> dict[str, NDArray]:
        """Read navigation data from the HKT products."""
        nav_data = {}
        for hkt_file in self.flnames:
            with h5py.File(hkt_file) as fid:
                # pylint: disable=no-member
                gid = fid["/navigation_data"]
                for key in gid:
                    data = gid[key][:]
                    if key.endswith("_time"):
                        data = np.datetime64(
                            gid["att_time"].attrs["units"].decode().split(" ")[2]
                        ) + (1e6 * data).astype("timedelta64[us]")

                    nav_data[key] = (
                        data
                        if key not in nav_data
                        else np.append(nav_data[key], data, axis=0)
                    )

        # order data in time
        if not np.all(nav_data["att_time"][:-1] <= nav_data["att_time"][1:]):
            indx = np.argsort(nav_data["att_time"])
            for key in nav_data:
                if key.startswith("att_"):
                    nav_data[key] = nav_data[key][indx, ...]

        if not np.all(nav_data["orb_time"][:-1] <= nav_data["orb_time"][1:]):
            indx = np.argsort(nav_data["orb_time"])
            for key in nav_data:
                if key.startswith("orb_"):
                    nav_data[key] = nav_data[key][indx, ...]

        if not np.all(nav_data["tilt_time"][:-1] <= nav_data["tilt_time"][1:]):
            indx = np.argsort(nav_data["tilt_time"])
            for key in nav_data:
                if key.startswith("tilt"):
                    nav_data[key] = nav_data[key][indx, ...]

        # set time_coverage
        self.time_coverage = (nav_data["att_time"][0], nav_data["att_time"][-1])

        return nav_data

    def nav_coverage_adjust(
        self: HKTio,
        nav_data: dict[str, NDArray],
        coverage_spx: tuple(np.datetime64, np.datetime64),
    ) -> dict[str, NDArray] | None:
        """Return navigation data within SPEXone time-coverage extended by 10 sec."""
        coverage = (coverage_spx[0] - TEN_SECONDS, coverage_spx[1] + TEN_SECONDS)

        att_mask = (nav_data["att_time"] >= coverage[0]) & (
            nav_data["att_time"] <= coverage[1]
        )
        orb_mask = (nav_data["orb_time"] >= coverage[0]) & (
            nav_data["orb_time"] <= coverage[1]
        )
        tilt_mask = (nav_data["tilt_time"] >= coverage[0]) & (
            nav_data["tilt_time"] <= coverage[1]
        )

        nav_dict = {}
        for key, value in nav_data.items():
            if key.startswith("att_"):
                nav_dict[key] = value[att_mask, ...]
            elif key.startswith("orb_"):
                nav_dict[key] = value[orb_mask, ...]
            else:
                nav_dict[key] = value[tilt_mask, ...]

        # update time_coverage
        self.time_coverage = (nav_data["att_time"][0], nav_data["att_time"][-1])

        return nav_dict

    def nav_coverage_flag(
        self: HKTio,
        coverage_spx: tuple(np.datetime64, np.datetime64),
    ) -> int:
        """Check completeness of the navigation data."""
        return CoverageFlag.check(self.time_coverage, coverage_spx)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2025 SRON
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""Tools to read or write definitions of SPEXone binning-tables."""

from __future__ import annotations

__all__ = ["BinningTables"]

import datetime as dt
from importlib.resources import files
from pathlib import Path
from typing import TYPE_CHECKING, Self

import h5py
import numpy as np

from .lib import pyspex_version

if TYPE_CHECKING:
    from numpy.typing import NDArray


# - global parameters ------------------------------
DATA_DIR_CSV = Path("/nfs/SPEXone/share/binning")
# DATA_DIR_CSV = Path("/data/richardh/SPEXone/share/binning")

# name of the output file with binning-table definitions
NAME_BIN_TBL = Path("./binning_tables.nc")


# - class BinningCKD -------------------------------
class BinningCKD:
    """Class to store the SPEXone binning-table definitions.

    Parameters
    ----------
    overwrite :  bool, default=False
       Any existing file may be overwritten

    Raises
    ------
    FileNotFoundError
        Directory with SPEXone binning-table CSV files does not exist.

    Notes
    -----
    The binning tables as defined on-ground are supposed to be available
    during the whole mission at the same on-board memory location.
    Because these original binning tables are necessary for re-processing
    and may facilitate instrument performance monitoring.
    Therefore, it is preferred that a new binning table is added to the
    current set, without changing the validity start string.
    However, new binning-table file should be released in case any of the
    binning tables are overwritten.

    Examples
    --------
    # create new file with binning-table definitions::

    > with BinningCKD(overwrite=True) with bin_ckd:
    >    bin_ckd.add_tbl_pre_launch()
    >    bin_ckd.add_tbl_post_launch()

    """

    def __init__(self: BinningCKD, overwrite: bool = False) -> None:
        """Initialize new file to store SPEXone binning-table."""
        self.fid = None
        if NAME_BIN_TBL.is_file() and not overwrite:
            print("Please use parameter 'overwrite' to replace existing CKD.")
            return

        fid = h5py.File(NAME_BIN_TBL, "w")
        fid.attrs["title"] = "SPEXone Level-1 binning-tables"
        fid.attrs["creator_name"] = "SRON/Earth"
        fid.attrs["creator_email"] = "SPEXone-MPC@sron.nl"
        fid.attrs["creator_url"] = "https://www.sron.nl/missions-earth/pace-spexone"
        fid.attrs["institution"] = "SRON"
        fid.attrs["publisher_name"] = "SRON/Earth"
        fid.attrs["publisher_email"] = "SPEXone-MPC@sron.nl"
        fid.attrs["publisher_url"] = "https://www.sron.nl/missions-earth/pace-spexone"
        fid.attrs["project"] = "PACE Project"
        fid.attrs["conventions"] = "CF-1.10 ACDD-1.3"
        fid.attrs["keyword_vocabulary"] = (
            "NASA Global Change Master Directory (GCMD) Science Keywords"
        )
        fid.attrs["stdname_vocabulary"] = (
            "NetCDF Climate and Forecast (CF) Metadata Convention"
        )
        fid.attrs["title"] = "PACE SPEXone Level-1A Data"
        fid.attrs["platform"] = "PACE"
        fid.attrs["instrument"] = "SPEXone"
        fid.processing_version = pyspex_version()
        fid.attrs["date_created"] = (
            dt.datetime.now(dt.UTC).replace(tzinfo=None).isoformat(timespec="seconds")
        )
        dset = fid.create_dataset("column", dtype="u2", shape=(1024,))
        dset.make_scale()
        dset = fid.create_dataset("row", dtype="u2", shape=(1024,))
        dset.make_scale()
        self.fid = fid

    def __enter__(self: BinningCKD) -> Self:
        """Initiate the context manager."""
        return self

    def __exit__(self: BinningCKD, *args: object) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: BinningCKD) -> None:
        """Close resources, after sanity check of SQLite and HDF5 databases."""
        if self.fid:
            # print(f"CLOSE HDF5 file: {self.fid.filename}")
            self.fid.close()

    @staticmethod
    def bin_tbl_defs() -> dict[int, tuple[str, str]]:
        """Return listing of binning-table definitions.

        Returns
        -------
        dict[int, tuple[str, str]]
            binning-table ID, names of files with lineskip and flexible binning

        """
        return {
            20: ("lineskiparrays_EXT_9p8_FM_0.csv", "binning_EXT_9p8_FM_0_add.csv"),
            24: ("lineskiparrays_EXT_9p8_FM_4.csv", "binning_EXT_9p8_FM_4_add.csv"),
            29: ("lineskiparrays_EXT_9p8_FM_8.csv", "binning_EXT_9p8_FM_8_add.csv"),
            50: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_1.csv"),
            51: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_2.csv"),
            52: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_3.csv"),
            53: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_4.csv"),
            54: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_5.csv"),
            55: ("lineskiparrays_170_offs2_2.csv", "binning_170_offs2_2_add_6.csv"),
            56: ("lineskiparrays_160_views.csv", "binning_160_views_add_1.csv"),
            57: ("lineskiparrays_160_views.csv", "binning_160_views_add_2.csv"),
            58: ("lineskiparrays_160_views.csv", "binning_160_views_add_3.csv"),
            59: ("lineskiparrays_160_views.csv", "binning_160_views_add_4.csv"),
            60: ("lineskiparrays_160_views.csv", "binning_160_views_add_5.csv"),
        }

    def __write__(self: BinningCKD, data_dir: Path, group: str | None = None) -> None:
        """Write binning-table definition to a HDF5 file.

        Parameters
        ----------
        data_dir :  Path
            Directory with the binning-table definitions
        group :  str, optional
            root in the HDF5 file where the data will be stored

        """
        for tbl_id, (line_fl, bin_fl) in self.bin_tbl_defs().items():
            lineskip_data = np.loadtxt(data_dir / line_fl, delimiter=",").astype("u1")
            binning_data = np.loadtxt(data_dir / bin_fl, delimiter=",", dtype="u4")
            iadd = (data_dir / bin_fl).stem.split("_")[-1]
            iadd = 1 if iadd == "add" else int(iadd)
            lineskip_data = lineskip_data[iadd, :]
            index, count = np.unique(
                binning_data[lineskip_data == 1, :], return_counts=True
            )
            gid = self.fid.create_group(
                str(Path("/" if group is None else group) / f"Table_{tbl_id:03d}")
            )
            gid.attrs["tabel_id"] = tbl_id
            gid.attrs["REG_BINNING_TABLE_START"] = hex(
                0x80000000 + 0x400000 * (tbl_id - 1)
            )
            gid.attrs["enabled_lines"] = np.uint16(lineskip_data.sum())
            gid.attrs["flex_binned_pixels"] = np.uint32(index.max() + 1)

            dset = gid.create_dataset("bins", dtype="u2", shape=(count.size,))
            dset.make_scale()

            dset = gid.create_dataset(
                "count_table",
                count.shape,
                dtype="u2",
                compression="gzip",
                compression_opts=1,
                shuffle=True,
            )
            dset.dims[0].attach_scale(gid["bins"])
            dset.attrs["long_name"] = "number of aggregated pixel readings"
            dset.attrs["valid_min"] = np.uint16(0)
            dset.attrs["valid_max"] = np.uint16(count.max())
            dset[:] = count.astype("u2")

            dset = gid.create_dataset(
                "lineskip_arr",
                lineskip_data.shape,
                dtype="u1",
                compression="gzip",
                compression_opts=1,
                shuffle=True,
            )
            dset.dims[0].attach_scale(self.fid["row"])
            dset.attrs["long_name"] = "lineskip array"
            dset.attrs["valid_min"] = np.uint8(0)
            dset.attrs["valid_max"] = np.uint8(1)
            dset[:] = lineskip_data

            dset = gid.create_dataset(
                "binning_table",
                binning_data.shape,
                dtype="u4",
                fillvalue=np.iinfo("u4").max,
                chunks=(128, 128),
                compression="gzip",
                compression_opts=1,
                shuffle=True,
            )
            dset.dims[0].attach_scale(self.fid["row"])
            dset.dims[1].attach_scale(self.fid["column"])
            dset.attrs["long_name"] = "binning table"
            dset.attrs["valid_min"] = np.uint32(0)
            dset.attrs["valid_max"] = np.uint32(index.max())
            dset[:] = binning_data

    def add_tbl_pre_launch(self: BinningCKD) -> None:
        """Add tables with flexible binning used before launch."""
        data_dir = DATA_DIR_CSV / "20210208T152000"
        if not data_dir.is_dir():
            raise FileNotFoundError(f"can not find folder: {data_dir}")
        gid = self.fid.create_group("20210208T152000")
        gid.attrs["validity_date"] = str(np.datetime64("2021-02-08T15:20:00"))
        self.__write__(data_dir, "20210208T152000")

    def add_tbl_post_launch(self: BinningCKD) -> None:
        """Add tables with flexible binning used after launch."""
        data_dir = DATA_DIR_CSV / "20210304T124000"
        if not data_dir.is_dir():
            raise FileNotFoundError(f"can not find folder: {data_dir}")
        self.fid.attrs["validity_date"] = str(np.datetime64("2021-03-04T12:40:00"))
        self.__write__(data_dir)


# - class BinningTables ----------------------------
class BinningTables:
    """Class to apply the SPEXone binning-tables.

    Parameters
    ----------
    table_id :  int, optional
       Binning table ID number between one and 255
    coverage_start :  np.datetime64, optional
       Default is to return the table definitions as used after 2021-03-04

    Examples
    --------
    # use binning-table '130' to unbin SPEXone detector data::

    > with BinningTables(130) as bin_tbl:
    >    img_2d = bin_tbl.to_image(img_binned)

    """

    def __init__(
        self: BinningTables,
        table_id: int | None = None,
        *,
        coverage_start: np.datetime64 | None = None,
    ) -> None:
        """Initialize class attributes."""
        binning_db = files("pyspex.Data").joinpath("binning_tables.nc")
        if table_id is None:
            return

        if not binning_db.is_file():
            raise FileNotFoundError(f"{binning_db} not found")

        with h5py.File(binning_db) as fid:
            if coverage_start is None or coverage_start > np.datetime64(
                "2021-03-04T12:40:00"
            ):
                ds_name = f"Table_{table_id:03d}"
            else:
                ds_name = f"/20210208T152000/Table_{table_id:03d}"
            if ds_name not in fid:
                raise KeyError(f"{ds_name} not defined")
            self.binning_table = fid[f"{ds_name}/binning_table"][:]
            self.lineskip_arr = fid[f"{ds_name}/lineskip_arr"][:]
            self.count_table = fid[f"{ds_name}/count_table"][:]

    def __enter__(self: BinningCKD) -> Self:
        """Initiate the context manager."""
        return self

    def __exit__(self: BinningCKD, *args: object) -> bool:
        """Exit the context manager."""
        return False  # any exception is raised by the with statement.

    def _unbin_(self: BinningTables, img_1d: NDArray) -> NDArray[float]:
        """Return detector data corrected for number of flexible binnings (still 1-D).

        Parameters
        ----------
        img_1d :  NDArray
           Binned image data (N * 1-D array) where N is number of images

        Returns
        -------
        NDArray[float]
           unbinned image data (N * 1-D array)

        Examples
        --------
        Let's read 2 executions of measurements with MPS=47 from a SPEXone L1A product:

          > spx = SPXtlm()
          > spx.from_l1a(l1a_product, tlm_type='sci', mps_id=47)
          > print(len(spx.science.images), spx.science.images[0].shape)
          2, (14, 203500)
          > bin_tbl = BinningTables(spx.science.binning_table())
          > img_1d = bin_tbl.unbin(spx.science.images)
          > print(img_1d.shape)
          (28, 203500)

        """
        img_1d = img_1d.astype(float)
        img_1d /= self.count_table
        mask = self.count_table > 5

        if img_1d.ndim == 2:
            img_1d[:, mask] = np.nan
            return img_1d

        img_1d[mask] = np.nan
        return img_1d

    def to_image(
        self: BinningTables, img_binned: NDArray | tuple[NDArray], unbin: bool = True
    ) -> NDArray:
        """Return 2x2 binned 2-D detector image data.

        Parameters
        ----------
        img_binned : NDArray | tuple[NDArray]
           image data (N * 1-D array)
        unbin :  bool, default=True
           return image data as floats corrected for flexible binning

        Returns
        -------
        NDArray
           2x2 binned image with shape=(N, 1024, 1024)

        """
        if isinstance(img_binned, tuple):
            img_1d = np.concatenate(img_binned, axis=0)
        else:
            img_1d = np.asarray(img_binned)

        if unbin:
            img_1d = self._unbin_(img_1d)

        fill_value = (
            np.nan
            if np.issubdtype(img_1d.dtype, np.floating)
            else np.iinfo(img_1d.dtype).max
        )
        table = self.binning_table[self.lineskip_arr == 1, :].reshape(-1)
        if img_1d.ndim == 2:
            new_shape = (img_1d.shape[0], *self.binning_table.shape)
            revert = np.full(new_shape, fill_value)
            revert[:, self.lineskip_arr == 1, :] = img_1d[:, table].reshape(
                img_1d.shape[0], -1, 1024
            )
        else:
            revert = np.full(self.binning_table.shape, fill_value)
            revert[self.lineskip_arr == 1, :] = img_1d[table].reshape(-1, 1024)

        return revert

    def to_binned(
        self: BinningTables,
        image: NDArray | tuple[NDArray],
    ) -> NDArray[np.uint16]:
        """Return a 1-D array binned according to the current binning-table.

        Parameters
        ----------
        image :  NDArray
           2-D detector image with shape (N, 1024, 1024) or (N, 2024, 2024)

        Returns
        -------
        NDarray[uint16]
           binned 1-D image data

        """
        if isinstance(image, tuple):
            img_2d = np.concatenate(image, axis=0)
        else:
            img_2d = np.asarray(image)
        n_img = img_2d.shape[0]

        # perform 2x2 binning
        if img_2d.shape[1] == img_2d.shape[2] == 2048:
            img_2d = np.sum(img_2d.reshape(1024, 2, 1024, 2), axis=(1, 3))

        # pylint: disable=no-member
        binning_table = self.binning_table.reshape(-1)

        # perform flexible binning
        buf_1d = np.full((n_img, self.count_table.size), np.nan)
        uvals, counts = np.unique(self.binning_table, return_counts=True)

        # start_time = time.time()
        mask = counts == 1
        mask2 = np.isin(binning_table, uvals[mask])
        buf_1d[:, uvals[mask]] = img_2d.reshape(n_img, -1)[:, mask2]
        # print(f"processing count==1 took: {time.time() - start_time:.3f} sec.")

        # start_time = time.time()
        mask = counts == 2
        mask2 = np.isin(binning_table, uvals[mask])
        buf_1d[:, uvals[mask]] = np.sum(
            img_2d.reshape(n_img, -1)[:, mask2].reshape(n_img, -1, 2), axis=2
        )
        # print(f"processing count==2 took: {time.time() - start_time:.3f} sec.")

        # start_time = time.time()
        mask = counts == 3
        # pylint: disable=no-member
        mask2 = np.isin(binning_table, uvals[mask])
        buf_1d[:, uvals[mask]] = np.sum(
            img_2d.reshape(n_img, -1)[:, mask2].reshape(n_img, -1, 3), axis=2
        )
        # print(f"processing count==3 took: {time.time() - start_time:.3f} sec.")

        # return array with data-type uint16
        buf_1d[np.isnan(buf_1d)] = np.iinfo("u2").max
        return buf_1d.astype("u2")
